from typing import List

from cement import Controller, ex

from banyan.api import PolicyAPI, AttachmentAPI
from banyan.controllers.base import Base
from banyan.model.policy import PolicyInfo, Policy


class PolicyController(Controller):
    class Meta:
        label = 'policy'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage authorization policies for users and workloads'

    @property
    def _client(self) -> PolicyAPI:
        return self.app.client.policies

    @ex(help='list policies')
    def list(self):
        policies: List[PolicyInfo] = self._client.list()
        results = list()
        headers = ['Name', 'ID', 'Created', 'Last Updated']
        for policy in policies:
            new_row = [policy.policy_name, policy.policy_id,
                       policy.created_at.strftime(Base.TABLE_DATE_FORMAT),
                       policy.last_updated_at.strftime(Base.TABLE_DATE_FORMAT)]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show the definition of a policy',
        arguments=[
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to display.'
             }),
        ])
    def get(self):
        info: PolicyInfo = self._client[self.app.pargs.policy_name]
        policy_json = Policy.Schema().dump(info.policy)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(policy_json, handler='json', indent=2, sort_keys=True)

    @ex(help='create a new policy from a JSON specification',
        arguments=[
            (['policy_spec'],
             {
                 'help': 'JSON blob describing the new policy to be created, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def create(self):
        policy_spec = Base.get_json_input(self.app.pargs.policy_spec)
        policy = Policy.Schema().load(policy_spec)
        policy_info = self._client.create(policy)
        self.app.render(PolicyInfo.Schema().dump(policy_info), handler='json', indent=2, sort_keys=True)

    @ex(help='update an existing policy',
        arguments=[
            (['policy_spec'],
             {
                 'help': 'JSON blob describing the policy to be updated, or a filename '
                         'containing JSON prefixed by "@" (example: @service.json).'
             }),
        ])
    def update(self):
        policy_spec = Base.get_json_input(self.app.pargs.policy_spec)
        policy = Policy.Schema().load(policy_spec)
        policy_info = self._client.update(policy)
        self.app.render(PolicyInfo.Schema().dump(policy_info), handler='json', indent=2, sort_keys=True)

    @ex(help='delete a policy',
        arguments=[
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to delete.'
             }),
        ])
    def delete(self):
        policy: PolicyInfo = self._client[self.app.pargs.policy_name]
        self.app.print(self._client.delete(policy))

    @ex(help='show services attached to a policy',
        arguments=[
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to delete.'
             }),
        ])
    def list_attachments(self):
        policy: PolicyInfo = self._client[self.app.pargs.policy_name]
        attach_api: AttachmentAPI = self.app.client.attachments
        attachments = attach_api.list()
        results = list()
        headers = ['Object Name', 'Object ID', 'Type', 'Attached At', 'Attached By']
        for a in attachments:
            if a.policy_id == policy.id:
                new_row = [a.attached_to_name, a.attached_to_id, a.attached_to_type,
                           a.attached_at.strftime(Base.TABLE_DATE_FORMAT), a.attached_by]
                results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='attach a policy to a service',
        arguments=[
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to attach to the service.',
             }),
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to attach a policy to.',
             }),
            (['--permissive'],
             {
                 'action': 'store_false',
                 'dest': 'enforcing',
                 'help': 'Set the policy to permissive mode (allow all traffic and log any unauthorized access).',
             }),
            (['--enforcing'], {
                'action': 'store_true',
                'dest': 'enforcing',
                'default': True,
                'help': 'Set the policy to enforcing mode (deny unauthorized access).',
            }),
        ])
    def attach_to_service(self):
        info = self._client.attach(self.app.pargs.policy_name, self.app.pargs.service_name, self.app.pargs.enforcing)
        mode = 'ENFORCING' if info.enabled else 'PERMISSIVE'
        self.app.print(f'Policy {info.policy_id} attached to service {info.service_id} in {mode} mode.')

    @ex(help='detach the policy from a service',
        arguments=[
            (['policy_name'],
             {
                 'metavar': 'policy_name_or_id',
                 'help': 'Name or ID of the policy to attach to the service.',
             }),
            (['service_name'],
             {
                 'metavar': 'service_name_or_id',
                 'help': 'Name or ID of the service to attach a policy to.',
             }),
        ])
    def detach_from_service(self):
        self.app.print(self._client.detach(self.app.pargs.policy_name, self.app.pargs.service_name))



import json
import sys

from cement import Controller
from cement.utils.version import get_version_banner

from ..core.version import get_version

VERSION_BANNER = """
API library and command-line interface for Banyan Security %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    TABLE_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    class Meta:
        label = 'base'
        title = 'Commands'

        usage = 'banyan [options] <command> <subcommand> [<subcommand> ...] [parameters]'
        # text displayed at the top of --help output
        description = 'API library and command-line interface for Banyan Security'

        # text displayed at the bottom of --help output
        # epilog = 'Usage: banyan command1 --foo bar'

        # controller level arguments. ex: 'banyan --version'
        arguments = [
            # add a version banner
            (['-v', '--version'],
             {'action': 'version',
              'version': VERSION_BANNER}),
            (['--api-url'],
             {'help': 'URL for the Banyan API server. Can also be configured via the BANYAN_API_URL '
                      'environment variable.'}),
            (['--refresh-token'],
             {'help': 'API token used for the initial authentication to the Banyan API server. Can also be '
                      'configured via the BANYAN_REFRESH_TOKEN environment variable.'}),
            (['--output-format', '-o'],
             {'choices': ['table', 'json', 'yaml'],
              'help': 'desired output format (table, json, yaml)'}),
        ]

    def _post_argument_parsing(self):
        super()._post_argument_parsing()
        if self.app.pargs.api_url:
            self.app.client.api_url = self.app.pargs.api_url
        if self.app.pargs.refresh_token:
            self.app.client.refresh_token = self.app.pargs.refresh_token

    def _default(self):
        """Default action if no sub-command is passed."""
        self.app.args.print_help()

    @staticmethod
    def get_json_input(arg: str):
        if arg[0] == '@':
            arg = open(arg[1:]).read()
        elif arg == '-':
            arg = sys.stdin.read()
        else:
            arg = arg.encode('utf-8')
        return json.loads(arg)



from dataclasses import field
from datetime import datetime
from ipaddress import IPv4Interface
from typing import List, Dict, Optional, ClassVar, Type
from uuid import UUID

from marshmallow import validate, fields, pre_load, Schema
from marshmallow_dataclass import dataclass
# from dataclasses_serialization.json import JSONStrSerializerMixin, JSONSerializerMixin
# from banyan.model.custom_types import Iso8601Date
from semver import VersionInfo

from banyan.model import Resource


@dataclass
class Netagent(Resource):
    TERMINATED = "Terminated"
    REPORTING = "Reporting"
    TAG_HOSTNAME = "com.banyanops.hosttag.hname"
    TAG_TYPE = "com.banyanops.hosttag.type"
    TAG_SITE_NAME = "com.banyanops.hosttag.site_name"
    TAG_DOMAIN_NAMES = "com.banyanops.hosttag.site_domain_names"
    TAG_FQDN = "com.banyanops.hosttag.site_address"
    hostname: str = field(metadata={"data_key": "Hostname"})
    version: VersionInfo = field(metadata={"marshmallow_field": fields.String(data_key="Version")})
    visibility: bool = field(metadata={"data_key": "Visibility"})
    cidrs_txt: str = field(metadata={"data_key": "CIDRs"})
    uname_txt: str = field(metadata={"data_key": "Uname"})
    site_name: str = field(metadata={"data_key": "SiteName"})
    status: str = field(metadata={"validate": validate.OneOf([TERMINATED, REPORTING]), "data_key": "Status"})
    cluster_name: Optional[str] = field(metadata={"data_key": "ClusterName"})
    cluster_id: Optional[UUID] = field(metadata={"data_key": "ClusterID"})
    last_activity_at: Optional[datetime] = field(metadata={"format": "iso", "data_key": "LastActivityAt"})
    ip_addresses: List[str] = field(default_factory=list, metadata={"data_key": "IPs"})
    host_tags: Dict[str, str] = field(default_factory=dict, metadata={"data_key": "HostTags"})
    Schema: ClassVar[Type[Schema]] = Schema

    # noinspection PyUnusedLocal
    @pre_load
    def _remove_empty_dates(self, data, many, **kwargs):
        if "LastActivityAt" in data and data["LastActivityAt"] == "":
            del data["LastActivityAt"]
        return data

    @property
    def name(self) -> str:
        return self.hostname

    @property
    def id(self) -> str:
        return self.hostname

    @property
    def domain_names(self) -> List[str]:
        names = self.host_tags.get(Netagent.TAG_DOMAIN_NAMES).split(',')
        return [x.strip() for x in names]

    @property
    def fqdn(self):
        return self.host_tags.get(Netagent.TAG_FQDN)

    @property
    def host_data(self) -> Dict[str, str]:
        data = dict()
        for pair in self.uname_txt.split(','):
            k, *v = pair.strip().split(' ')
            data[k] = ' '.join(v)
        return data

    @property
    def cidr_map(self) -> Dict[IPv4Interface, int]:
        assert(self.cidrs_txt[0] == '[' and self.cidrs_txt[-1] == ']')
        tokens = self.cidrs_txt[1:-1].split(' ')
        cidrs = dict()
        while tokens:
            net_str = IPv4Interface(tokens[0])
            assert(tokens[1] == '--dport')
            port = int(tokens[2])
            tokens = tokens[3:]
            cidrs[net_str] = port
        return cidrs

    @property
    def public_ip_addr(self) -> str:
        for ip in self.ip_addresses:
            nic = IPv4Interface(ip)
            if nic.is_global:
                return ip
        return ''

    @property
    def is_access_tier(self):
        host_type = self.host_tags.get(Netagent.TAG_TYPE)
        return host_type == 'access_tier'

    @property
    def host_type(self):
        return self.host_tags.get(Netagent.TAG_TYPE)



from cement import Controller, ex


class EventV1Controller(Controller):
    class Meta:
        label = 'event'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'report on security and audit events'

    # @property
    # def _client(self) -> RoleAPI:
    #     return self.app.client.roles

    # TODO: implement /security_events
    @ex(help='list events')
    def list(self):
        pass

    # TODO: implement /security_events_type_count
    @ex(help='show summary of events in database')
    def summary(self):
        pass



import dns.resolver
from banyan.model.netagent import Netagent


n: Netagent = Netagent.Schema().loads(open("tests/data/netagent.json").read())

resolver = dns.resolver.Resolver(configure=False)
resolver.nameservers = ['1.1.1.1']
answers = resolver.resolve('pst.transtar.bnndemos.com', 'A')
for answer in answers:  # type: dns.resolver.Answer
    print(answer.address, answer.address in n.ip_addresses)



from cement import Controller, ex


class AdminController(Controller):
    class Meta:
        label = 'admin'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage administrator accounts'

    # @property
    # def _client(self) -> RoleAPI:
    #     return self.app.client.roles

    # TODO: implement
    @ex(help='list administrators')
    def list(self):
        pass

    # TODO: implement /admin/add_user
    @ex(help='create an administrator')
    def create(self):
        pass

    # TODO: implement /admin/delete_user
    @ex(help='delete an administrator')
    def delete(self):
        pass

    # TODO: implement /admin/update_user
    @ex(help='update settings for an existing administrator')
    def update(self):
        pass

    # TODO: implement /admin/saml
    @ex(help='get SAML SSO configuration for administrator login')
    def saml(self):
        pass

    # TODO: implement /admin/setup_saml
    @ex(help='set SAML SSO configuration for administrator login')
    def setup_saml(self):
        pass



from dataclasses import field
from typing import List, Dict, Optional, Union
from uuid import UUID

from marshmallow import validate
from marshmallow_dataclass import dataclass

from banyan.model import BanyanApiObject, InfoBase


@dataclass
class Tags:
    TEMPLATE_USER = "USER"
    TEMPLATE_WORKLOAD = "WORKLOAD"
    TEMPLATE_CUSTOM = "CUSTOM"
    _TEMPLATE_VALUES = (TEMPLATE_USER, TEMPLATE_WORKLOAD, TEMPLATE_CUSTOM)
    template: str = field(metadata={"validate": validate.OneOf(_TEMPLATE_VALUES)})


@dataclass
class Metadata:
    name: str
    description: str
    tags: Tags


@dataclass
class Spec:
    EMPLOYEE_OWNED = "E"
    CORPORATE_DEDICATED = "D"
    CORPORATE_SHARED = "C"
    OTHER = "O"
    _OWNERSHIP_VALUES = (EMPLOYEE_OWNED, CORPORATE_DEDICATED, CORPORATE_SHARED, OTHER)
    known_device_only: Optional[bool]
    label_selector: List[Dict[str, str]] = field(default_factory=list)
    service_account: List[str] = field(default_factory=list)
    image: List[str] = field(default_factory=list)
    repo_tag: List[str] = field(default_factory=list)
    container_fqdn: List[str] = field(default_factory=list)
    group: List[str] = field(default_factory=list)
    email: List[str] = field(default_factory=list)
    device_ownership: List[str] = field(default_factory=list)


@dataclass
class Role(BanyanApiObject):
    KIND = "BanyanRole"
    metadata: Metadata
    spec: Spec

    def __post_init__(self):
        self.kind = self.KIND

    @property
    def name(self):
        return self.metadata.name


@dataclass
class RoleInfo(InfoBase):
    role_id: UUID = field(metadata={'data_key': 'RoleID'})
    role_name: str = field(metadata={'data_key': 'RoleName'})
    spec: str = field(metadata={'data_key': 'RoleSpec'})
    description: str = field(metadata={'data_key': 'Description'})

    type: str = field(metadata={'data_key': 'RoleType'})
    version: int = field(metadata={'data_key': 'RoleVersion'})
    enabled: bool = field(metadata={'data_key': 'Enabled'})

    @property
    def role(self) -> Role:
        return Role.Schema().loads(self.spec)

    @property
    def name(self) -> str:
        return self.role_name

    @property
    def id(self) -> str:
        return str(self.role_id)


RoleInfoOrName = Union[RoleInfo, str]



from typing import List

from cement import Controller, ex

from banyan.api import NetagentAPI, ShieldAPI
from banyan.controllers.base import Base
from banyan.model.netagent import Netagent


class NetagentController(Controller):
    class Meta:
        label = 'netagent'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage netagents (AccessTiers and HostAgents)'

    @property
    def _client(self) -> NetagentAPI:
        return self.app.client.netagents

    @ex(help='list netagents',
        arguments=[
            (['--all'],
             {
                 'action': 'store_true',
                 'dest': 'show_all',
                 'help': 'Display all agents, including terminated agents. The default is to show only '
                         'active, reporting agents.',
             }),
        ])
    def list(self):
        agents: List[Netagent] = self._client.list()
        clusters: ShieldAPI = self.app.client.shields
        results = list()
        headers = ['Hostname', 'Shield', 'Type', 'Public IP', 'Status', 'Last Activity']
        for agent in agents:
            shield_name = agent.cluster_name or clusters.find(str(agent.cluster_id)).name
            if agent.status == 'Reporting' or self.app.pargs.show_all:
                last_activity = agent.last_activity_at.strftime(Base.TABLE_DATE_FORMAT) if \
                    agent.last_activity_at else 'None'
                new_row = [agent.hostname, shield_name, agent.host_type,
                           agent.public_ip_addr, agent.status,
                           last_activity]
                results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show detailed information about a netagent',
        arguments=[
            (['hostname'],
             {
                 'help': 'Hostname of the agent.'
             }),
        ])
    def get(self):
        hostname = self.app.pargs.hostname
        agent = self._client.find(hostname)
        agent_json = Netagent.Schema().dump(agent)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(agent_json, handler='json', indent=2, sort_keys=True)



from dataclasses import field
from datetime import datetime
from typing import List, ClassVar, Type, Optional
from uuid import UUID

from marshmallow import Schema, validate, fields, pre_load
from marshmallow_dataclass import dataclass
from semver import VersionInfo

from banyan.model import NanoTimestampField, Resource
from banyan.model.trustscore import TrustFactorsV2


@dataclass
class TrustFactorV1:
    name: str = field(metadata={'data_key': 'Name'})
    value: bool = field(metadata={'data_key': 'Value'})
    type: str = field(metadata={'data_key': 'Type'})
    source: str = field(metadata={'data_key': 'Source'})
    description: str = field(metadata={'data_key': 'Description'})


@dataclass
class TrustDataV1:
    ALWAYS_ALLOW = "AlwaysAllow"
    ALWAYS_DENY = "AlwaysDeny"
    LOW = "Low"
    MEDIUM = "Medium"
    HIGH = "High"
    TRUST_VALUES = (ALWAYS_DENY, LOW, MEDIUM, HIGH, ALWAYS_ALLOW)
    entity_trustscore: int = field(metadata={'data_key': 'EntityTrustscore'})
    override_trustscore: int = field(metadata={'data_key': 'OverrideTrustscore'})
    access_trustscore: int = field(metadata={'data_key': 'AccessTrustscore'})
    override_active: bool = field(metadata={'data_key': 'OverrideActive'})
    level: str = field(metadata={'data_key': 'Level', 'validate': validate.OneOf(TRUST_VALUES)})
    updated_at: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='UpdatedAt')})
    factors: List[TrustFactorV1] = field(default_factory=list, metadata={'data_key': 'Factors'})


@dataclass
class MdmData:
    timestamp: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='Timestamp')})
    source: str = field(metadata={'data_key': 'Source'})
    compromised_status: str = field(metadata={'data_key': 'CompromisedStatus'})
    compliant_status: str = field(metadata={'data_key': 'CompliantStatus'})


@dataclass
class User(Resource):
    username: str = field(metadata={'data_key': 'Name'})
    email: str = field(metadata={'data_key': 'Email'})
    groups: str = field(metadata={'data_key': 'Groups'})
    last_login: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='LastLogin')})
    login_count: int = field(metadata={'data_key': 'LoginCount'})
    trust_data: TrustDataV1 = field(metadata={'data_key': 'TrustData'})
    serial_numbers: List[str] = field(default_factory=list, metadata={'data_key': 'SerialNumbers'})
    roles: List[str] = field(default_factory=list, metadata={'data_key': 'Roles'})
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.username

    @property
    def id(self) -> str:
        return self.email


@dataclass
class Device(Resource):
    CORPORATE_DEDICATED = 'Corporate Dedicated'
    CORPORATE_SHARED = 'Corporate Shared'
    EMPLOYEE_OWNED = 'Employee Owned'
    OTHER = 'Other'
    _OWNERSHIP_VALUES = (CORPORATE_DEDICATED, CORPORATE_SHARED, EMPLOYEE_OWNED, OTHER)

    serial_number: str = field(metadata={'data_key': 'SerialNumber'})
    device_friendly_name: str = field(metadata={'data_key': 'DeviceFriendlyName'})
    last_login: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='LastLogin')})
    login_count: int = field(metadata={'data_key': 'LoginCount'})
    ownership: str = field(metadata={'data_key': 'Ownership', 'validate': validate.OneOf(_OWNERSHIP_VALUES)})
    platform: str = field(metadata={'data_key': 'Platform'})
    model: str = field(metadata={'data_key': 'Model'})
    architecture: str = field(metadata={'data_key': 'Architecture'})
    registered_status: bool = field(metadata={'data_key': 'RegisteredStatus'})
    is_banned: bool = field(metadata={'data_key': 'Banned'})
    os_level: str = field(metadata={'data_key': 'OS'})
    app_version: Optional[VersionInfo] = field(metadata={'marshmallow_field': fields.String(data_key='AppVersion'),
                                                         "missing": None})
    mdm_data: MdmData = field(metadata={'data_key': 'MdmData'})
    trust_data: TrustDataV1 = field(metadata={'data_key': 'TrustData'})
    emails: List[str] = field(default_factory=list, metadata={'data_key': 'Emails'})
    roles: List[str] = field(default_factory=list, metadata={'data_key': 'Roles'})
    device_id: UUID = field(default=None, metadata={'data_key': 'DeviceID', 'missing': ''})
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.device_friendly_name

    @property
    def id(self) -> str:
        return self.serial_number

    # noinspection PyUnusedLocal
    @pre_load
    def _remove_empty_id(self, data, **kwargs):
        if 'DeviceID' in data and data['DeviceID'] == '':
            del data['DeviceID']
        return data


@dataclass
class TrustScore(Resource):
    TRUST_TYPE_DEVICE = 'Device'
    TRUST_TYPE_USER = 'EndUser'
    TRUST_TYPE_EXTERNAL = 'External'
    _TRUST_TYPE_VALUES = (TRUST_TYPE_DEVICE, TRUST_TYPE_USER, TRUST_TYPE_EXTERNAL)
    trust_type: str = field(metadata={'data_key': 'TrustType', 'validate': validate.OneOf(_TRUST_TYPE_VALUES)})
    trust_id: str = field(metadata={'data_key': 'TrustID'})
    score: int = field(metadata={'data_key': 'Score'})
    level: str = field(metadata={'data_key': 'Level'})
    ext_source: str = field(metadata={'data_key': 'ExtSource'})
    reason: str = field(metadata={'data_key': 'Reason'})
    factors_json: str = field(metadata={'data_key': 'FactorsJSON'})
    created_at: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='CreatedAt')})
    deleted_at: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='DeletedAt')})
    last_updated_at: datetime = field(metadata={"marshmallow_field": NanoTimestampField(data_key='LastUpdatedAt')})
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.trust_id

    @property
    def id(self) -> str:
        return self.trust_id

    @property
    def factors(self) -> TrustFactorsV2:
        return TrustFactorsV2.Schema().loads(self.factors_json)



from dataclasses import field
from ipaddress import IPv4Interface
from typing import List, Dict, Union, Optional, ClassVar

from marshmallow import validate, pre_load, fields, Schema
from marshmallow_dataclass import dataclass

from banyan.model import BanyanApiObject, InfoBase, IPv4InterfaceField


@dataclass
class Tags:
    TEMPLATE_WEB = "USER_WEB"
    TEMPLATE_TCP = "USER_TCP"
    TEMPLATE_CUSTOM = "CUSTOM"
    APP_TYPE_WEB = "WEB"
    APP_TYPE_SSH = "SSH"
    APP_TYPE_TCP = "TCP"
    APP_TYPE_RDP = "RDP"
    APP_TYPE_GENERIC = "GENERIC"
    protocol: str
    domain: str
    user_facing: bool = field(metadata={'marshmallow_field': fields.String()})
    template: str
    service_app_type: str
    icon: str = field(default="")
    port: int = field(default=443, metadata={'marshmallow_field': fields.String()})
    Schema: ClassVar[Schema] = Schema

    # noinspection PyUnusedLocal
    @pre_load
    def _handle_empty_port(self, data, **kwargs):
        if 'port' in data and data['port'] == '':
            del data['port']
        return data


@dataclass
class Metadata:
    tags: Tags
    name: str
    description: str
    cluster: str


@dataclass
class FrontendAddress:
    port: int = field(metadata={'marshmallow_field': fields.String()})
    cidr: IPv4Interface = field(metadata={'marshmallow_field': IPv4InterfaceField()})


@dataclass
class Attributes:
    tls_sni: List[str] = field(default_factory=list)
    frontend_addresses: List[FrontendAddress] = field(default_factory=list)
    host_tag_selector: List[Dict[str, str]] = field(default_factory=list)
    addresses: List[str] = field(default_factory=list)  # Deprecated


@dataclass
class CustomTLSCert:
    enabled: bool
    cert_file: str = ""
    key_file: str = ""


@dataclass
class CertSettings:
    custom_tls_cert: CustomTLSCert = field(default_factory=CustomTLSCert)
    dns_names: List[str] = field(default_factory=list)


@dataclass
class OIDCSettings:
    enabled: bool
    service_domain_name: str
    post_auth_redirect_path: str
    api_path: str
    trust_callbacks: Dict[str, str] = field(default_factory=dict)


@dataclass
class HTTPRedirect:
    enabled: bool
    url: str
    status_code: int
    addresses: List[str] = field(default_factory=list)
    from_address: List[str] = field(default_factory=list)


@dataclass
class HTTPHealthCheck:
    METHOD_GET = "GET"
    METHOD_HEAD = "HEAD"
    METHOD_DELETE = "DELETE"
    METHOD_PUT = "PUT"
    METHOD_POST = "POST"
    METHOD_OPTIONS = "OPTIONS"
    METHOD_ALL = "*"
    _METHOD_VALUES = (METHOD_GET, METHOD_POST, METHOD_HEAD, METHOD_PUT, METHOD_DELETE, METHOD_OPTIONS, METHOD_ALL, "")
    enabled: bool
    path: str
    user_agent: str
    https: bool
    method: str = field(metadata={"validate": validate.OneOf(_METHOD_VALUES)})
    addresses: List[str] = field(default_factory=list)
    from_address: List[str] = field(default_factory=list)


@dataclass
class Host:
    origin_header: List[str] = field(default_factory=list)
    target: List[str] = field(default_factory=list)


@dataclass
class Pattern:
    hosts: List[Host] = field(default_factory=list)
    methods: List[str] = field(default_factory=list)
    mandatory_headers: List[str] = field(default_factory=list)
    paths: List[str] = field(default_factory=list)


@dataclass
class ExemptedPaths:
    enabled: bool
    paths: Optional[List[str]]
    patterns: List[Pattern] = field(default_factory=list)


@dataclass
class HttpSettings:
    enabled: bool
    oidc_settings: Optional[OIDCSettings]
    http_health_check: Optional[HTTPHealthCheck]
    http_redirect: Optional[HTTPRedirect]
    exempted_paths: Optional[ExemptedPaths]


@dataclass
class CIDRAddress:
    cidr: IPv4Interface = field(metadata={'marshmallow_field': IPv4InterfaceField()})
    ports: int


@dataclass
class ClientCIDRs:
    addresses: List[CIDRAddress] = field(default_factory=list)
    host_tag_selector: List[Dict[str, str]] = field(default_factory=list)
    clusters: List[str] = field(default_factory=list)


@dataclass
class BackendTarget:
    name: str
    name_delimiter: Optional[str]
    port: int = field(default=443, metadata={'marshmallow_field': fields.String()})
    tls: Optional[bool] = True
    tls_insecure: Optional[bool] = False
    client_certificate: Optional[bool] = False

    # noinspection PyUnusedLocal
    @pre_load
    def _handle_empty_port(self, data, **kwargs):
        if 'port' in data and data['port'] == '':
            del data['port']
        return data


@dataclass
class Backend:
    target: BackendTarget
    dns_overrides: Dict[str, str] = field(default_factory=dict)
    whitelist: List[str] = field(default_factory=list)


@dataclass
class Spec:
    attributes: Attributes
    backend: Backend
    cert_settings: CertSettings
    http_settings: HttpSettings
    client_cidrs: List[ClientCIDRs] = field(default_factory=list)


@dataclass
class Service(BanyanApiObject):
    metadata: Metadata
    spec: Spec

    @property
    def name(self):
        return self.metadata.name


@dataclass
class OIDCClient:
    trust_auth: str
    trust_cb: str
    client_id: str
    client_secret: str
    trust_callbacks: Dict[str, str] = field(default_factory=dict)


@dataclass
class ServiceInfo(InfoBase):
    service_id: str = field(metadata={'data_key': 'ServiceID'})
    service_name: str = field(metadata={"data_key": 'ServiceName'})
    description: str = field(metadata={"data_key": 'Description'})
    spec: str = field(metadata={"data_key": 'ServiceSpec'})

    discovery: str = field(metadata={'data_key': 'ServiceDiscovery'})
    cluster_name: str = field(metadata={'data_key': 'ClusterName'})
    type: str = field(metadata={'data_key': 'ServiceType'})
    version: int = field(metadata={'data_key': 'ServiceVersion'})
    external: str = field(metadata={'data_key': 'External'})
    oidc_enabled: bool = field(metadata={'data_key': 'OIDCEnabled'})
    oidc_client_spec: str = field(metadata={'data_key': 'OIDCClientSpec'})
    user_facing: bool = field(metadata={'data_key': 'UserFacing'})
    protocol: str = field(metadata={'data_key': 'Protocol'})
    domain: str = field(metadata={'data_key': 'Domain'})
    port: int = field(metadata={'data_key': 'Port'})
    enabled: bool = field(metadata={'data_key': 'Enabled'})
    Schema: ClassVar[Schema] = Schema

    @property
    def service(self) -> Service:
        return Service.Schema().loads(self.spec)

    @property
    def oidc_client(self) -> OIDCClient:
        return OIDCClient.Schema().loads(self.oidc_client_spec)

    @property
    def name(self) -> str:
        return self.service_name

    @property
    def id(self) -> str:
        return self.service_id


ServiceInfoOrName = Union[ServiceInfo, str]



from dataclasses import field
from datetime import datetime
from typing import ClassVar, Type
from uuid import UUID

from marshmallow import Schema, validate
from marshmallow_dataclass import dataclass

from banyan.model import NanoTimestampField, Resource


@dataclass
class Attachment(Resource):
    SERVICE = "service"
    SAAS_APP = "saasapp"
    _ATTACH_TYPES = (SERVICE, SAAS_APP)

    policy_id: UUID = field(metadata={'data_key': 'PolicyID'})
    policy_name: str = field(metadata={'data_key': 'PolicyName'})
    attached_to_id: str = field(metadata={'data_key': 'AttachedToID'})
    attached_to_name: str = field(metadata={'data_key': 'AttachedToName'})
    attached_to_type: str = field(metadata={'data_key': 'AttachedToType',
                                            'validate': validate.OneOf(_ATTACH_TYPES)})
    enabled: bool = field(metadata={'data_key': 'Enabled'})
    attached_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField(data_key='AttachedAt')})
    attached_by: str = field(metadata={'data_key': 'AttachedBy'})
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.policy_name + '|' + self.attached_to_type + '|' + self.attached_to_name

    @property
    def id(self) -> str:
        return str(self.policy_id) + '|' + self.attached_to_type + '|' + self.attached_to_id



from typing import List

from cement import Controller, ex

from banyan.api import UserAPI
from banyan.controllers.base import Base
from banyan.model.user_device import User, TrustDataV1, TrustScore


class UserController(Controller):
    class Meta:
        label = 'user'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'manage user accounts'

    @property
    def _client(self) -> UserAPI:
        return self.app.client.users

    @ex(help='list users')
    def list(self):
        users: List[User] = self._client.list()
        results = list()
        headers = ['Name', 'Email', 'Last Login', 'Login Count', 'Trust Score', 'Last TS Update']
        for user in users:
            new_row = [user.name, user.email, user.last_login.strftime(Base.TABLE_DATE_FORMAT),
                       user.login_count, user.trust_data.level,
                       user.trust_data.updated_at.strftime(Base.TABLE_DATE_FORMAT)]
            results.append(new_row)
        results.sort(key=lambda x: x[0])
        self.app.render(results, handler='tabulate', headers=headers, tablefmt='simple')

    @ex(help='show detailed information about a user',
        arguments=[
            (['email'],
             {
                 'help': 'Email address for the user.'
             }),
        ])
    def get(self):
        email = self.app.pargs.email
        user = self._client.find(email)
        user_json = User.Schema().dump(user)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(user_json, handler='json', indent=2, sort_keys=True)

    @ex(help='show details about the user\'s trust score calculation',
        arguments=[
            (['email'],
             {
                 'help': 'Email address for the user.'
             }),
        ])
    def get_trust(self):
        email = self.app.pargs.email
        user = self._client.find(email)
        trust = self._client.get_trustscores(user)
        trust_json = TrustScore.Schema().dump(trust, many=True)
        # colorized_json = highlight(policy_json, lexers.JsonLexer(), formatters.Terminal256Formatter(style="default"))
        self.app.render(trust_json, handler='json', indent=2, sort_keys=True)

    @ex(help='set the maximum trust level for a user',
        arguments=[
            (['email'],
             {
                 'help': 'Email address for the user.'
             }),
            (['--trust-level'],
             {
                 'choices': TrustDataV1.TRUST_VALUES,
                 'required': True,
                 'help': f'Maximum trust level for this device. Must be one of: {TrustDataV1.TRUST_VALUES}.'
             }
             ),
            (['--reason'],
             {
                 'required': True,
                 'help': 'Explanation to be displayed in console and to the end user.'
             }),
            (['--ext-source'],
             {
                 'required': True,
                 'help': 'Name of the external data source (e.g. CarbonBlack, CrowdStrike, etc).'
             }),
        ])
    def set_max_trust(self):
        serial_number = self.app.pargs.serial_number
        device = self._client.find(serial_number)
        trust = self._client.set_max_trustlevel(device, self.app.pargs.trust_level,
                                                self.app.pargs.reason, self.app.pargs.ext_source)
        trust_json = TrustScore.Schema().dump(trust)
        self.app.render(trust_json, handler='json', indent=2, sort_keys=True)

    # TODO: implement /user/verify
    @ex(help='force verify a user')
    def verify(self):
        pass

    # TODO: implement /forgot_password
    @ex(help='send password reminder email')
    def forgot_password(self):
        pass

    # TODO: implement /reset_password
    @ex(help='send password reset email')
    def reset_password(self):
        pass

    # TODO: implement /refresh_token
    @ex(help='generate API refresh token for a user')
    def refresh_token(self):
        pass

    # TODO: implement /revoke_token
    @ex(help='revoke a user\'s refresh token')
    def revoke_token(self):
        pass



from typing import List

from banyan.model.attachment import Attachment


class AttachmentAPI:
    def __init__(self, client):
        self._client = client

    def list(self) -> List[Attachment]:
        response_json = self._client.api_request('GET', '/policy/attachment')
        return Attachment.Schema().load(response_json, many=True)



import logging
import os
from typing import Dict, Any

import requests
from requests.auth import AuthBase

from banyan.api.attachment import AttachmentAPI
from banyan.api.device import DeviceAPI
from banyan.api.netagent import NetagentAPI
from banyan.api.policy import PolicyAPI
from banyan.api.role import RoleAPI
from banyan.api.service import ServiceAPI
from banyan.api.shield import ShieldAPI
from banyan.api.user import UserAPI
from banyan.core.exc import BanyanError


class BanyanApiClient:
    DEFAULT_API_URL = 'https://net.banyanops.com'
    JSON_TYPE = 'application/json'

    class _Auth(AuthBase):
        def __init__(self, parent) -> None:
            self._obj = parent

        def __call__(self, r: requests.Request) -> requests.Request:
            token = self._obj.access_token or self._obj.refresh_token
            r.headers['Authorization'] = f'Bearer {token}'
            return r

    def __init__(self, api_server_url: str = None, refresh_token: str = None, debug: bool = False,
                 log: logging.Logger = None) -> None:
        self._debug = debug
        self._log = log
        self._api_url = self._normalize_url(api_server_url or BanyanApiClient.DEFAULT_API_URL)
        self._access_token = None
        self._refresh_token = refresh_token
        if not self._refresh_token:
            raise BanyanError("Refresh token must be set")
        if self._debug:
            requests_log = logging.getLogger('requests.packages.urllib3')
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
        self._http = self._create_session()
        self._services = ServiceAPI(self)
        self._policies = PolicyAPI(self)
        self._attach = AttachmentAPI(self)
        self._roles = RoleAPI(self)
        self._shields = ShieldAPI(self)
        self._agents = NetagentAPI(self)
        self._users = UserAPI(self)
        self._devices = DeviceAPI(self)

    # noinspection PyMethodMayBeStatic
    def _normalize_url(self, url: str) -> str:
        if '/api' not in url:
            url += '/api/v1'
        # if not url.endswith('/'):
        #     url += '/'
        return url

    def _create_session(self) -> requests.Session:
        http = requests.session()
        http.headers['Accept'] = 'application/json'
        http.auth = self._Auth(self)
        if self._debug:
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
        return http

    def get_access_token(self) -> str:
        content = self._request('POST', '/refresh_token').json()
        self._access_token = content['Message']
        return self._access_token

    def _request(self, method: str, url: str, params: Dict[str, str] = None, data: Any = None,
                 headers: Dict[str, str] = None, cookies: Dict[str, str] = None,
                 files=None, auth=None, timeout=None, allow_redirects=True, proxies=None,
                 hooks=None, stream=None, verify=None, cert=None, json=None) -> requests.Response:
        if '://' not in url:
            url = self._api_url + url
        response = self._http.request(method, url, params, data, headers, cookies, files, auth or self._http.auth,
                                      timeout, allow_redirects, proxies, hooks, stream, verify, cert, json)
        if response.status_code >= 400:
            try:
                content = response.json()
                if 'Message' in content:
                    raise BanyanError(f'{response.status_code} Client Error: {response.reason} for '
                                      f'url: {response.url}: {content["Message"]}')
            except ValueError:
                raise BanyanError(f'{response.status_code} Client Error: {response.reason} for url: {response.url}')
        return response

    def api_request(self, method: str, uri: str, params: Dict[str, str] = None, data: Any = None,
                    json: Any = None, headers: Dict[str, str] = None, accept: str = None) -> requests.Response:
        if not self._access_token:
            self.get_access_token()
        headers = headers or dict()
        headers['Accept'] = accept or self.JSON_TYPE
        if data:
            headers['Content-Type'] = self.JSON_TYPE
        return self._request(method=method, url=uri, params=params, data=data, headers=headers, json=json).json()

    @property
    def access_token(self) -> str:
        return self._access_token

    @property
    def refresh_token(self) -> str:
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, value: str) -> None:
        self._refresh_token = value
        if self._access_token:
            self._access_token = None
            self.get_access_token()

    @property
    def api_url(self) -> str:
        return self._api_url

    @api_url.setter
    def api_url(self, value: str) -> None:
        self._api_url = self._normalize_url(value)

    @property
    def services(self) -> ServiceAPI:
        return self._services

    @property
    def policies(self) -> PolicyAPI:
        return self._policies

    @property
    def roles(self) -> RoleAPI:
        return self._roles

    @property
    def attachments(self) -> AttachmentAPI:
        return self._attach

    @property
    def shields(self) -> ShieldAPI:
        return self._shields

    @property
    def netagents(self) -> NetagentAPI:
        return self._agents

    @property
    def users(self):
        return self._users

    @property
    def devices(self):
        return self._devices


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    c = BanyanApiClient(api_server_url='https://gcstage.banyanops.com', refresh_token=os.getenv('BANYAN_REFRESH_TOKEN'),
                        debug=True)
    print(c.get_access_token())
    print(c.services.list())




from setuptools import setup, find_packages
from banyan.core.version import get_version

VERSION = get_version()

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='pybanyan',
    version=VERSION,
    description='API library and command-line interface for Banyan Security',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Todd Radel',
    author_email='todd@banyansecurity.io',
    url='https://github.com/banyansecurity/pybanyan/',
    license='apache',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'banyan': ['templates/*']},
    include_package_data=True,
    entry_points={'console_scripts': [
        "banyan = banyan.main:main"
    ]},
)



from banyan.api.base import ServiceBase
from banyan.model.policy import PolicyInfo, Policy, PolicyInfoOrName, PolicyAttachInfo
from banyan.model.service import ServiceInfoOrName


class PolicyAPI(ServiceBase):
    class Meta:
        data_class = Policy
        info_class = PolicyInfo
        arg_type = PolicyInfoOrName
        list_uri = '/security_policies'
        delete_uri = '/delete_security_policy'
        insert_uri = '/insert_security_policy'
        uri_param = 'PolicyID'
        obj_name = 'policy'

    def attach(self, policy: PolicyInfoOrName, service: ServiceInfoOrName, enforcing: bool) -> PolicyAttachInfo:
        from banyan.api.service import ServiceAPI
        policy = self.find(policy)
        service = ServiceAPI(self._client).find(service)
        json_response = self._client.api_request('POST', '/insert_security_attach_policy',
                                                 params={
                                                     'PolicyID': policy.id,
                                                     'ServiceID': service.id,
                                                     'Enabled': str(enforcing).upper()
                                                 })
        return PolicyAttachInfo.Schema().load(json_response)

    def detach(self, policy: PolicyInfoOrName, service: ServiceInfoOrName) -> str:
        from banyan.api.service import ServiceAPI
        policy = self.find(policy)
        service = ServiceAPI(self._client).find(service)
        json_response = self._client.api_request('DELETE', '/delete_security_attach_policy',
                                                 params={
                                                     'PolicyID': policy.id,
                                                     'ServiceID': service.id
                                                 })
        return json_response['Message']
